//
// Adaptive Prediction Trees
// Image Coding
//
// Version 1.0
//
// Based on...
//
// Binary tree predictive coding versions 1-5
//
// John Robinson 1994, 1995, 1997, 2002, 2003, 2004
// john@userport.com
//

// Declarations (and some definitions) of internals.

#include "apt.h"
// #define DEBUG 1

// The first section is for the compactor classes

namespace apt1_0 {

// #define DEBUG 1
// #define NOCOMPACTORDEBUG 1

const int TABLEBITS = 12;
const int MASK = (1<<TABLEBITS)-1;
const int TABLESIZE = (1<<TABLEBITS);

class compactor {
private:
	int debug;

	int id;
	class cfileio *pfio;
	int longest;
	int longmask;
	unsigned char num_symbols_of_length[32];
	unsigned char ordered_symbols[256];
	long accum;
	int maxval;
	int minval;
	int litlength;
	int usinghuffman;	// 1 if using Huffman, 0 if literals
// Next the stuff for encoding
	long freq[256];
	unsigned short encode[256];
	unsigned char enclength[256];
// Then the stuff for decoding
	unsigned char decode[TABLESIZE];
	unsigned char declength[TABLESIZE];
	int usetree(int omit_literal_table) const;
	void buildtree(int f[], unsigned char num_symbols_of_length[],
		unsigned char ordered_symbols[], int& longest) const;
	void buildbothtrees();
public:
	compactor(){
		debug = 0;
		id = 0;
		pfio = 0; 
		longest = 0;
		reset();
		}
	~compactor() {}
	void putbit(int val);
	int getbit();
	void reset();
	void debugon() { debug = 1; }
	void debugoff() { debug = 0; }
	void indexnumber(int i) { id = i; }
	void attach(class cfileio *f) { pfio = f; }
	void inc(int symbol) {freq[symbol]++;};
	void dec(int symbol) {freq[symbol]--;};
	void writetree(void);
	void maketable_encode(void);
	void put(int inbyte);
	void incorput(int symbol, int which) {
		if (which)
			put(symbol);
		else
			freq[symbol]++;
		}
	void readtree(void);
	void maketable_decode(void);
	int get();
	void copytree(compactor& source);
	int copytreewithout(compactor& source, int from, int to);
	};

class cfileio {
private:
	FILE	*cfp;
// Output works a byte at a time and is therefore simple
	int	outbyte;
	int	outbits; // Number of output bits
// Input is double buffered: first get a chunk into buf, then get bits into
// bitbuf
	int charsleft;
	unsigned long	bitbuf;
	int valid;	// bits left in bitbuf
	unsigned char *pbuf;
	unsigned char buf[2048];
	int fillbuf() {
		return(charsleft = fread((pbuf = buf),1,2048,cfp));
		}
	void loadforlook() {
		bitbuf = (bitbuf<<8) + (charsleft-- ? *pbuf++ : 
			(fillbuf() && charsleft--) ? *pbuf++ : 0);
		}
	int lookbits(const int& bits, const int& mask) {
		if (valid < bits)
			{
			loadforlook();
			valid += 8;
			}
		if (valid < bits)
			{
			loadforlook();
			valid += 8;
			}
		return (int) ((bitbuf >> (valid-bits)) & mask);
		}
	void skipbits(const int bits) {
		valid -= bits;
		}
	int get(unsigned char *declength, unsigned char *decode, 
	    register int bits, const int mask) {
		register int v = valid;
		if (v < bits) {
			loadforlook();
			if ((v += 8) < bits) {
				loadforlook();
				v += 8;
				}
			}
		bits = (int) ((bitbuf >> (v-bits)) & mask);
		valid = v - declength[bits];
		return decode[bits];
		}
public:
	cfileio(FILE *file) { 
		cfp = file; outbyte = 0; outbits = 0;
		bitbuf = 0; charsleft = 0; valid = 0; }
	void putbit(int val) {
		outbyte <<= 1;
		if (val)
			outbyte++;
		outbits++;
		if ((outbits &= 0x7) == 0)
			putc(outbyte, cfp);
		}
	void putbits(int val, int numbits) {
		if (!numbits)
			return;
		int space = 8-outbits;
		if (numbits > space) {
			outbyte <<= space;
			outbyte |= val>>(numbits = numbits-space);
			putc(outbyte, cfp);
			outbits = 0;
			space = 8;
			val &= (1<<numbits) - 1;
			if (numbits > space) { 
			// Now space = 8 and outbyte is empty
			    outbyte = val>>(numbits = numbits-space);
			    putc(outbyte, cfp);
			    val &= (1<<numbits) - 1;
			    }
			}
		// Now guaranteed that numbits <= space
		outbyte <<= numbits;
		outbyte |= val;
		outbits += numbits;
		if ((outbits &= 0x7) == 0)
			putc(outbyte, cfp);
		}
	void putbyte(unsigned char val) {
		int thisbyte = 8 - outbits;
		outbyte <<= thisbyte;
		outbyte |= (val >> outbits);
		putc(outbyte, cfp);
		outbyte = val;
		}
	void flush() {
		if (outbits) {
			outbyte <<= 8 - outbits;
			putc(outbyte ,cfp);
			outbits = 0;
			}
		fflush(cfp);
		}
	int getbit() {
		if (!valid) {
			bitbuf = (charsleft-- ? *pbuf++ : 
				(fillbuf() && charsleft--) ? *pbuf++ : 0);
			valid = 8;
			}
		return (int) ((bitbuf >> --valid) & 1);
		}
	int getbits(int numbits) {
		// numbits <= 8
		int mask = (1<<numbits) - 1;
		int retval = lookbits(numbits, mask);
		skipbits(numbits);
		return retval;
		}
	int rewind() {
		int retval = -charsleft-(valid>>3);
		fseek(cfp, -charsleft-(valid>>3), SEEK_CUR);
		charsleft = 0;
		valid = 0;
		return retval;
		}
	~cfileio() {}
friend class compactor;
	};

inline void compactor::putbit(int val) {
#ifdef DEBUG
	if (debug)
		printf("Coder %d: bit %d\n",id,val);
#endif
	pfio->putbit(val);
	}
inline int compactor::getbit() {
	int code = pfio->getbit();
#ifdef DEBUG
	if (debug)
		printf("Coder %d: bit %d\n",id,code);
#endif
	return code;
	}
inline int compactor::get ()
{
	int code;
	if (!usinghuffman) {
		code = pfio->getbits(litlength);
#ifdef DEBUG
	if (debug) {
		printf("Coder %d: uncoded symbol %d (0x%x)\n",id,code,code);
		fflush(stdout);
		}
#endif
		return code;
		}
	code = pfio->get(declength, decode, longest, longmask);
#ifdef DEBUG
	if (debug) {
		printf("Coder %d: symbol %d (0x%x)\n",id,code,code);
		fflush(stdout);
		}
#endif
	return code;
}

const int FLAGWORDCHUNK = 3;
const int FLAGWORDS = 3*3*3;
const int MAXFLAG = 63;
// Runlength/huffman compression class
class rlcomp {
private:
	int debug;
	int id;
	int usingflagwords;	// Flag to say whether using flagwords
	int symbolcnt;
	int doublecode;		// Two successive flagwords of all freqsymbol
	unsigned char freqsymbol;
	unsigned char secondfreqsymbol;
	unsigned char decodeflagword;
	unsigned char *inbuf;
	unsigned char *procbuf;
	unsigned char *readproc;
	unsigned char *inbufend;
	unsigned char *procbufend;
	int buflength;
	compactor *psymbols;
	compactor *pflagwords;
	long freq[256];
	void disable_compactor_debug() {
		if (!debug)
			return;
		psymbols->debugoff();
		pflagwords->debugoff();
		}
	void enable_compactor_debug() {
		if (!debug)
			return;
		psymbols->debugon();
		pflagwords->debugon();
		}
public:
	rlcomp();
	~rlcomp();
	int encodelength(int maxlength); // Called to set up encode buffers
	void putbit(int val) { psymbols->putbit(val);}
	int getbit() { return psymbols->getbit();}
	void reset();
	void debugon() { 
		debug=1; psymbols->debugon(); pflagwords->debugon();}
	void debugoff() { 
		debug=0; psymbols->debugoff(); pflagwords->debugoff();}
	void indexnumber(int i) { 
		id = i;
		psymbols->indexnumber(i+100); 
		pflagwords->indexnumber(i+200);
		}
	void attach(class cfileio *f) { 
		psymbols->attach(f); pflagwords->attach(f);
		}
	void inc(int symbol) {
		freq[symbol]++; 
		*inbufend++ = (unsigned char) symbol;
		}
	void buildtree(void);
	void writetree(void);
	void readtree(void);
	void maketable_encode(void);
	void maketable_decode(void);
	void incorput(int symbol, int which);
	void put(int inbyte);
	int get(){
		if (!usingflagwords)
			return psymbols->get();
		if (!symbolcnt) {
			if (doublecode) {
				doublecode--;
				decodeflagword = FLAGWORDS - 1;
				}
			else {
				decodeflagword = pflagwords->get();
				if (decodeflagword >= FLAGWORDS) {
					doublecode = decodeflagword+1
							-FLAGWORDS;
					decodeflagword = FLAGWORDS - 1;
					}
				}
			}
		else
			decodeflagword *= 3;
		if (++symbolcnt == FLAGWORDCHUNK)
			symbolcnt = 0;
		int temp = decodeflagword/(FLAGWORDS/3);
		temp %= 3;
		if (temp == 2)
			return freqsymbol;
		if (temp)
			return secondfreqsymbol;
		return psymbols->get();
		}
	};

const int numcoders = 8;
const int SST = 0; // First numbered coder for square band
const int DST = 4; // First numbered coder for diamond band

inline int get_ridgethresh(int spacing) {
	// return (int)(10.0*log((double)spacing));
	// math-library-free approx to above:
	int ret = 2*spacing + 5;
	if (ret > 16)
		ret = 16 + (ret-16)/2;
	if (ret > 32)
		ret = 32 + (ret-32)/8;
	if (ret < 8) // Spacing = 1
		ret = 1;
	return ret;
	}
int get_numlevels();

int makescratch(const int& xmax, const int& ymax);

int deletescratch(const int& ymax);

void enc_plane(unsigned char **in, const int& xmax, const int& ymax,
const int& fromlevel, const int& tolevel, const int *spacing, 
rlcomp *codeout, const int& minval, const int& maxval, const int& comp,
const int& truecols, const int& truerows);

int bincode(unsigned char ***pyr, const int& inrows, const int& incols, 
rlcomp *codeout, int fromlevel, const int *spacing, 
const int pnmtype, const int& truerows, const int& truecols);

void calc_colour_transform(unsigned char ***components, int *transform, 
int num_rows, int num_cols);

void rgb_pca(unsigned char ***components, int *transform, 
int num_rows, int num_cols);

int calcq(int dynamicrange, int quality, int *qvals, int numbands);

void gettrueminmax(unsigned char ***components, const int pnmtype,
int *minval, int *maxval, const int& rows, const int& cols);

void pad(unsigned char ***incomponents, unsigned char ***pbuf, 
int& rows, int& cols, int pnmtype);

const unsigned char **initclip(const int& bottomval, const int& topval, 
	int pnmtype);

void dec_plane(unsigned char ***out,const int& XMAX, const int& YMAX,
const int& fromlevel, const int *spacing,
rlcomp *codein, const int *minval, const int *maxval,
const int pnmtype, const int& truecols, const int& truerows);

void pca_rgb(unsigned char ***components, int *transform, 
int num_rows, int num_cols, const unsigned char **clip);

int read_header(FILE *cfp, int& pnmtype, int& rows, int& cols, 
int *minval, int *maxval, int *tminval, int *tmaxval, int& numlevels, 
char& coltype, int *spacing, int *transform, char **comment);

};
