//
// Adaptive Prediction Trees version 1.0 encoder
//
// Converts raw PPM and PGM format files to APT encoded files using the
// function apt_encode()
//
// Copyright (c) John A Robinson. 1994, 1995, 1997, 2002, 2003, 2004
//
#include <cstdlib>
#include <cstring>
#include <cmath>
#include "apt.h"
using apt1_0::apt_encode;

//
// Helper functions for dealing with PPM/PGM files
//

void pgmword(FILE *in, char *cbuf) {
// The PNM man pages say you have to be as lenient as possible about pgm/ppm/pbm
// files, so this is my general-purpose lenient word reader.
// Read next PGM/PPM/PBM word, ignoring comments and whitespace
	char c;
	while(1) {
	    do 
		c = fgetc(in);
	    while(c == ' ' || c == '\t' || c == '\n' || c == '\r');
	    if (c == '#') {
		do
			c = fgetc(in);
		while (c != '\n');
		}
	    else break;
	    }
	*cbuf++ = c;
	c = fgetc(in);
	while (c != ' ' && c != '\t' && c != '\n' && c != '\r' && c != '#') {
	    *cbuf++ = c;
	    c = fgetc(in);
	    }
	fseek(in, -1, SEEK_CUR);
	*cbuf = '\0';
	}
FILE *readpnmheader(const char *pnmname, int& rows, int& cols, int& comps,
		int& numbits) {
	int maxval;
	char cbuf[32];
	unsigned char c;
	comps = -1;	// Signal for failure
	FILE *in=fopen(pnmname, "rb");
	if (!in) {
		fprintf(stderr, "Cannot open %s\n", pnmname);
		return 0;
		}
	cbuf[0] = fgetc(in);
	cbuf[1] = fgetc(in);
	cbuf[2] = 0;
	if (!strcmp(cbuf,"P5")) 
		// PGM image
		comps = 1;
	else if (!strcmp(cbuf,"P4"))
		// PBM image
		comps = 0;	// Used to signal pbm image
	else if (!strcmp(cbuf,"P6"))
		// PPM image
		comps = 3;
	else {
		fprintf(stderr,"Sorry. This version of capt only works with ");
		fprintf(stderr,"raw format (P4, P5 or P6) PBM/PGM/PPM files.\n");
		fprintf(stderr,"%s is not a valid raw format file\n",pnmname);
		return 0;
		}
	pgmword(in, cbuf);
	cols = atoi(cbuf);
	pgmword(in,cbuf);
	rows = atoi(cbuf);
	if (comps > 0) {
		pgmword(in,cbuf);
		maxval = atoi(cbuf);
		// We're going to throw this maxval away, but first...
		for (numbits = 0; numbits < 16; numbits++) {
			if (maxval < (1<<numbits))
				break;
			}
		}
	else
		numbits = 1;
	// absorb final whitespace char before start of data
	c = fgetc(in);
	return in;
	}

int readpnm(FILE *in, int& ymax, int& xmax, 
	unsigned char ***pbuf, int pnmtype, int numbits) {
// Read picture in "P4", "P5" or "P6" format into arrays.
	int i,j;
	int numbytes = 1;
	if (numbits > 8)
		numbytes = 2;
	if (pnmtype == 0)	// PBM: We map white to 255 and black to 0.
				// Later processing in encode detects bilevel
	    {
	    unsigned char **buf;
	    unsigned char *tempbuf = new unsigned char[(xmax+7)/8];
	    *pbuf++ = buf = new unsigned char *[ymax];
	    *pbuf++ = buf;
	    *pbuf = buf;	// So all three components point at same
	    for (i = 0; i < ymax; i++) {
		buf[i] = new unsigned char[xmax];
		fread(tempbuf, 1, (xmax+7)/8, in);
		unsigned char *tb = tempbuf;
		int bitmask = 0x80;
		for (j = 0; j < xmax; j++) {
			buf[i][j] = (*tb & bitmask) ? 0 : 255;
			bitmask >>= 1;
			if (!bitmask) {
				bitmask = 0x80;
				tb++;
				}
			}
		}
	    delete [] tempbuf;
	    }
	else if (pnmtype == 1)	// PGM 
	    {
	    unsigned char **buf;
	    *pbuf++ = buf = new unsigned char *[ymax];
	    *pbuf++ = buf;
	    *pbuf = buf;	// So all three components point at same
	    for (i = 0; i < ymax; i++) {
		buf[i] = new unsigned char[xmax*numbytes];
		fread(buf[i], numbytes, xmax, in);
		}
	    }
	else if (pnmtype == 3)	// PPM 
	    {
	    unsigned char **bufr, **bufg, **bufb;
	    unsigned char *tempbuf = new unsigned char[3*xmax*numbytes];
	    unsigned char *pr, *pg, *pb, *ptemp;
	    *pbuf++ = bufg = new unsigned char *[ymax];
	    *pbuf++ = bufr = new unsigned char *[ymax];
	    *pbuf = bufb = new unsigned char *[ymax];
	    for (i = 0; i < ymax; i++){
		pg = bufg[i] = new unsigned char[xmax*numbytes];
		pr = bufr[i] = new unsigned char[xmax*numbytes];
		pb = bufb[i] = new unsigned char[xmax*numbytes];
		fread(tempbuf,numbytes,xmax*3,in);
		ptemp = tempbuf;
		if (numbytes == 2) {	// 2 bytes per pel
		    for (j = 0; j < xmax; j++) {
			*pr++ = *ptemp++;
			*pr++ = *ptemp++;
			*pg++ = *ptemp++;
			*pg++ = *ptemp++;
			*pb++ = *ptemp++;
			*pb++ = *ptemp++;
			}
		    }
		else {
		    for (j = 0; j < xmax; j++) {
			*pr++ = *ptemp++;
			*pg++ = *ptemp++;
			*pb++ = *ptemp++;
			}
		    }
		}
	    delete [] tempbuf;
	    }
	fclose(in);
	return pnmtype;
	}

int main(int argc, const char *argv[])
	{
	if ((argc != 3)&&(argc != 4))
		{
		fprintf(stderr,"Usage: capt pnm_inpic code_file [quality]\n");
		return -1;
		}
	unsigned char **components[3];
	// char comment[] = "pnm to apt by apt1.0";
	char comment[] = "";
	int rows, cols;
	int pnmtype;
	int numbits;
	FILE *infile = readpnmheader(argv[1], rows, cols, pnmtype,
		numbits);
	if (pnmtype < 0)
		return -1;
	readpnm(infile,rows,cols, components, pnmtype, numbits);
	int quality = 80;
	if (argc == 4)
		quality = atoi(argv[3]);
	if (pnmtype == 0) {
		quality = 100;	// Always code at full quality
		pnmtype = 1;	// Just represent as PGM now
		}
	int retval;
	FILE *cfp = fopen(argv[2],"wb");
	if (cfp == 0) 
		retval = -2;
	else
		retval = apt_encode(components, cfp, rows, cols,
		pnmtype, quality, comment, numbits);
	switch (retval) {
		case -1:
			fprintf(stderr,"Quality parameter must be between 0 and 100\n");
			break;
		case -2:
			fprintf(stderr,"Can't open %s\n", argv[2]);
			break;
		default:
			// Success
			break;
		}
	for (int compnum = 0; compnum < pnmtype; compnum++) {
		unsigned char **in = components[compnum];
	   	for (int i = 0; i < rows; i++)
			delete [] in[i];
	    	delete [] in;
	    	}
	fclose(cfp);
	return retval;
	}

