//
// Adaptive Prediction Trees
// Image Coding
//
// Version 1.0
//
// Based on...
//
// Binary tree predictive coding versions 1-5
//
// John Robinson 1994, 1995, 1997, 2002, 2003, 2004
// john@userport.com
//

//
// Member functions for the compact data type implement a static Huffman coder.
//
#include <cstring>
#include "apt_internal.h"
using namespace apt1_0;

#define DEBUG 1
// #define DEBUG2 1

void compactor::reset(void) {
#ifdef DEBUG
	if (debug)
		printf("\nCoder %d: reset:\n",id); fflush(stdout);
#endif
	for (int i = 0; i < 256; i++)
		freq[i] = 0;
	}

void compactor::buildtree(int freq[], unsigned char num_symbols_of_length[],
		unsigned char ordered_symbols[], int& longest) const
// Based on JPEG huffman algorithm
	{
#ifdef DEBUG
	if (debug)
		printf("\nCoder %d: buildtree:\n",id); fflush(stdout);
#endif
	int i;
	int h1;
	int h2;
	int length[256];
	int descendents[256];
	long f[256];
	for (i = 0; i < 256; i++) {
		f[i] = freq[i];
		length[i] = 0;
		descendents[i] = -1;
		}
	while(1) {
		h1 = -1;
		h2 = -1;
		// Find smallest non-zero frequency
		for (i = 0; i < 256; i++) {
			if (f[i]) {
			    if (h1 == -1 || f[i] < f[h1]) {
				if (h2 == -1 || f[h1] < f[h2])
					h2 = h1;
				h1 = i;
			    } else if (h2 == -1 || f[i] < f[h2])
				h2 = i;
			}
		}
		if (h2 == -1) // Found the root
			break;
		// Merge the two trees
		f[h1] += f[h2];
		f[h2] = 0;
		length[h1]++;
		while (descendents[h1] >= 0) {
			h1 = descendents[h1];	// Go down chain
			length[h1]++;
			}
		//  Put h2 on this chain
		descendents[h1] = h2;
		// And go down its chain
		while (descendents[h1] >= 0) {
			h1 = descendents[h1];
			length[h1]++;
			}
		}
	if (h1 == -1) // Empty tree - create dummy char
		{
		longest = 0;
		num_symbols_of_length[0] = 1;
		ordered_symbols[0] = 0;
		return;
		}
	if (length[h1] == 0)	// one symbol
		{
		longest = 0;
		num_symbols_of_length[0] = 1;
		ordered_symbols[0] = h1;
		return;
		}
	for (i = 0; i < 32; i++)
		num_symbols_of_length[i] = 0;
	for (i = 0; i < 256; i++) {
		if (length[i])
			num_symbols_of_length[length[i]]++;
			// Assume never have case of 256 equal symbols
		}
	// Reduce to TABLEBITS-sized words using JPEG algorithm
	for (i = 31; i > TABLEBITS; i--) {
		while(num_symbols_of_length[i]) {
			int j = i - 2;
			while (!num_symbols_of_length[j])
				j--;
			num_symbols_of_length[i] -= 2;
			num_symbols_of_length[i-1]++;
			num_symbols_of_length[j+1] += 2;
			num_symbols_of_length[j]--;
			}
		}
	while(num_symbols_of_length[i] == 0)
		i--;
	// i is now length of longest codeword
	longest = i;
	int k = 0;
	for (i = 1; i <= 31; i++) {
		for (int j = 0; j < 256; j++)
			if (length[j] == i)
				ordered_symbols[k++] = j;
		}
	}

// Decide whether to use Huffman coding or not.
int compactor::usetree(int omit_literal_table) const {
#ifdef DEBUG
	if (debug) {
		printf("\nCoder %d: usetree(%d):\n",id,
			omit_literal_table); fflush(stdout);
		printf("Total symbols to code = %ld\n", accum);
		printf("longest = %d\n", longest); fflush(stdout);
		int k = 0;
		for (int i = 1; i <= longest; i++) {
			printf("num of length %d = %d\n",i, 
						num_symbols_of_length[i]);
			for (int j = 0; j < num_symbols_of_length[i]; j++, k++)
				printf("%d[n=%ld] ", ordered_symbols[k],
						freq[ordered_symbols[k]]);
			printf("\n");
			}
		}
#endif
	int litbits;
	int retval;
	int codelength = 0;
	int totallength;
	int totallitbits;
	int tablelength = 5+longest*8; // Because have to write 4 bits in any
				     // case at the start to give length of
				     // literals plus flag to say whether
				     // omitting table, then longest _bytes_ to
				     // say how many codewords there are of
				     // each length up to longest.
	// Work out the length of literals
	if (maxval > 127) litbits = 8;
	else if (maxval > 63) litbits = 7;
	else if (maxval > 31) litbits = 6;
	else if (maxval > 15) litbits = 5;
	else if (maxval > 7) litbits = 4;
	else if (maxval > 3) litbits = 3;
	else if (maxval > 1) litbits = 2;
	else if (maxval > 0) litbits = 1;
	else litbits = 0;
	// Then work out the total codelength
	int k = 0;
	for (int i = 0; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			{
			codelength += i*freq[ordered_symbols[k]];
			if (litbits) {
				k++;
				if (!omit_literal_table)
					tablelength += litbits;
				}
			}
		}
	totallength = codelength + tablelength;
	totallitbits = litbits*accum;
	retval = totallitbits - totallength;	// -ve means don't use Huffman
#ifdef DEBUG
    if (debug) {
	printf("Huffman code will code %ld symbols in %d bits (%d bytes).\n",
		accum, codelength, (codelength+7)/8);
	printf("To transmit code table required %d bits (%d bytes).\n",
		tablelength, (tablelength+7)/8);
	printf("Total bits required (table+code) = %d (%d bytes).\n",
		totallength, (totallength+7)/8);
	printf("Average rate = %f bits/symbol\n",(float)totallength/accum);
	printf("Maximum input symbol value is %d.\n", maxval);
	printf("Sending %d-bit literals directly requires %ld bits (%ld bytes).\n",
		litbits, litbits*accum, (litbits*accum+7)/8);
	if (totallitbits > totallength)
		printf("++Huffman coding gives saving of %d bits (%d bytes).\n",
			totallitbits-totallength, (totallitbits-totallength+7)/8);
	else
		printf("--Huffman coding adds additional %d bits (%d bytes).\n",
			totallength-totallitbits, (totallength-totallitbits+7)/8);
	printf("\n");
        }
#endif
	return retval;
	}

void compactor::buildbothtrees() {
// New compactor function to build Huffman codes based on actual freqs and
// on best match for constraint that increasing input values have increasing
// length codewords
	unsigned char c1_num_symbols_of_length[32];
	unsigned char c1_ordered_symbols[256];
	int c1_longest;
	int c1_usetree;
	unsigned char c2_num_symbols_of_length[32];
	unsigned char c2_ordered_symbols[256];
	int c2_longest;
	int c2_usetree;
	int f[256];
	int i;
	accum = 0;
	maxval = 0;
	minval = -1;
	for (i = 0; i < 256; i++) {
		f[i] = freq[i];
		accum += freq[i];
		if (freq[i] > 0) {
			maxval = i;
			if (minval < 0)
				minval = i;
			}
		}
	if (accum < 1) {
		accum = 1;
		minval = 0;
		}
	buildtree(f,c1_num_symbols_of_length,c1_ordered_symbols,c1_longest);
	for (i = 0; i < 32; i++)
		num_symbols_of_length[i] = c1_num_symbols_of_length[i];
	for (i = 0; i < 256; i++)
		ordered_symbols[i] = c1_ordered_symbols[i];
	longest = c1_longest;
	c1_usetree = usetree(0);	// Arg 0 means include literal table
	// Only consider c2 if minval can be represented in 2 bits
	// Also, if maxval is very high, it's certainly not worth doing this
	// and we have to avoid minval=0, maxval=255 case which could cause
	// errors in buildtree.
	if ((minval > 3)||(maxval > 200)) {
		if (c1_usetree <= 0)
			usinghuffman = 0;
		else
			usinghuffman = 1;
		return;
		}
	// Now manipulate things for c2
	for (i = minval; i <= maxval; i++)
		f[i] <<= 8;	// Multiply each count by 256
	for (i = maxval; i >= minval; i--) {
		// Have to ensure that these counts are in ascending order,
		// as if the least valued input had happened most times
		long cnt = f[i];
		for (int j = i-1; j >= minval; j--) {
			if (f[j] <= cnt)
				f[j] = cnt+1;
			}
		}
	buildtree(f,c2_num_symbols_of_length,c2_ordered_symbols,c2_longest);
	for (i = 0; i < 32; i++)
		num_symbols_of_length[i] = c2_num_symbols_of_length[i];
	for (i = 0; i < 256; i++)
		ordered_symbols[i] = c2_ordered_symbols[i];
	longest = c2_longest;
	c2_usetree = usetree(1);	// Arg 1 means omit literal table
	if ((c1_usetree <= 0)&&(c2_usetree <= 0))
		usinghuffman = 0;
	else if (c1_usetree > c2_usetree) {
		usinghuffman = 1;
		for (i = 0; i < 32; i++)
			num_symbols_of_length[i] = c1_num_symbols_of_length[i];
		for (i = 0; i < 256; i++)
			ordered_symbols[i] = c1_ordered_symbols[i];
		longest = c1_longest;
		}
	else
		usinghuffman = 2;
	}
	

// Write the Huffman tree to the output code file
void compactor::writetree(void) {
	buildbothtrees();
#ifdef DEBUG
	if (debug) {
		printf("\nCoder %d: writetree:\n",id); fflush(stdout);
		printf("Total symbols to code = %ld\n", accum);
		if (usinghuffman == 2)
			printf("Using Ordered Pseudo-Huffman:\n");
		else if (usinghuffman == 1)
			printf("Using Standard Huffman:\n");
		else
			printf("Not Using Huffman:\n");
		}
#endif
	pfio->putbit(usinghuffman);
	int litbits;
	int i;
	if (usinghuffman) {
		pfio->putbit(longest & 0x8);
		pfio->putbit(longest & 0x4);
		pfio->putbit(longest & 0x2);
		pfio->putbit(longest & 0x1);
		pfio->putbit(usinghuffman & 0x2);
		if (usinghuffman == 2) {
			pfio->putbit(minval & 0x2);
			pfio->putbit(minval & 0x1);
			}
		}
	// Work out the longest literal
	if (maxval > 127) litbits = 8;
	else if (maxval > 63) litbits = 7;
	else if (maxval > 31) litbits = 6;
	else if (maxval > 15) litbits = 5;
	else if (maxval > 7) litbits = 4;
	else if (maxval > 3) litbits = 3;
	else if (maxval > 1) litbits = 2;
	else if (maxval > 0) litbits = 1;
	else litbits = 0;
	litlength = litbits;
	pfio->putbit(litbits & 0x8);
	pfio->putbit(litbits & 0x4);
	pfio->putbit(litbits & 0x2);
	pfio->putbit(litbits & 0x1);
#ifdef DEBUG
	if (debug) {
		printf("litbits = %d\n", litbits); fflush(stdout);
		}
#endif
	if (!usinghuffman)
		return;
	for (i = 1; i <= longest; i++)
		pfio->putbyte(num_symbols_of_length[i]);
	if (usinghuffman == 1)
	    {
	    // the literals
	    int k = 0;
	    for (i = 0; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			{
			if (litbits)
				pfio->putbits(ordered_symbols[k++],litbits);
			}
		}
	    }
	fflush(stdout);
	maketable_encode();
	}

// Read the Huffman tree from the input code file.
void compactor::readtree(void) {
	int i;
	int startval = 0;
	usinghuffman = pfio->getbit();
	if (usinghuffman) {
		longest = pfio->lookbits(4,0xf);
		pfio->skipbits(4);
		usinghuffman += pfio->getbit();
		if (usinghuffman == 2) {
			startval = pfio->lookbits(2,0x3);
			pfio->skipbits(2);
			}
		}
	int litbits = pfio->lookbits(4,0xf);
	int litmask = (1<<litbits) - 1;
	pfio->skipbits(4);
	litlength = litbits;
#ifdef DEBUG
    if (debug) {
	printf("\nCoder %d: readtree:\n",id);
	if (usinghuffman == 2)
		printf("Ordered Pseudo-Huffman: longest codeword = %d\n", longest);
	else if (usinghuffman)
		printf("Standard Huffman: longest codeword = %d\n", longest);
	else
		printf("Not using Huffman\n");
	printf("litbits = %d\n", litbits); fflush(stdout);
	}
#endif
	if (!usinghuffman)
		return;
	if (!longest) {
		num_symbols_of_length[0] = 1;
		ordered_symbols[0] = 0;
		}
	else
		num_symbols_of_length[0] = 0;
	for (i = 1; i <= longest; i++) {
		num_symbols_of_length[i] = pfio->lookbits(8,0xff);
		pfio->skipbits(8);
#ifdef DEBUG
		if (debug)
			printf("num symbols of length %d = %d\n",i,num_symbols_of_length[i]);
#endif
		}
	int k = 0;
	for (i = 0; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++) {
			if (usinghuffman == 2) {
				ordered_symbols[k++] = startval++;
				}
			else {
				ordered_symbols[k++] = 
					pfio->lookbits(litbits,litmask);
				pfio->skipbits(litbits);
				}
#ifdef DEBUG
			if (debug)
				printf("%d ", ordered_symbols[k-1]);
#endif
			}
		}
#ifdef DEBUG
	if (debug) {
		printf("\n"); fflush(stdout);
		}
#endif
	maketable_decode();
	}
void compactor::maketable_encode(void) {
	// Create coding tables
	// First create temporary tables ordered by code length
#ifdef DEBUG
	// printf("(maketable_encode called by writetree:)\n", id);
#endif
	unsigned char codelength[257];
	unsigned short codebits[256];
	int k = 0;
	for (int i = 1; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			codelength[k++] = i;
		}
	codelength[k] = 0;	// Sentinal
	int lastsymbol = k;
	int codeword = 0;
	k = 0;
	int total = codelength[0];
	while(codelength[k]) {
		while(codelength[k] == total) {
			codebits[k++] = codeword;
			codeword++;
			}
		codeword <<= 1;
		total++;
		}
	// And finally install in lookup table
	memset(enclength,0,256);	// So all unused slots will be 0
	if (!longest) {
#ifdef DEBUG
		// printf("0-bit symbol: %d\n",ordered_symbols[0]); fflush(stdout);
#endif
		}
	for (k = 0; k < lastsymbol; k++) {
		encode[ordered_symbols[k]] = codebits[k];
		enclength[ordered_symbols[k]] = codelength[k];
#ifdef DEBUG
/*
		printf("%d: ",ordered_symbols[k]);
		for (int temp = 1; temp <= codelength[k]; temp++) {
			if (codebits[k] & (1<<(codelength[k]-temp)))
				printf("1");
			else
				printf("0");
			}
		printf("\n"); fflush(stdout);
*/
#endif
		}
	}
void compactor::maketable_decode(void) {
	// Create coding tables
#ifdef DEBUG
	if (debug) {
		printf("(maketable_decode called by readtree):\n"); fflush(stdout);
		}
#endif
	int tablesize = 1<<longest;
	longmask = tablesize - 1;
	if (!longest)	// Null table
		{
		memset(declength,0,TABLESIZE);
		memset(decode,ordered_symbols[0],TABLESIZE);
#ifdef DEBUG
		if (debug) {
			printf("Null table with single 0-bit symbol: %d\n",ordered_symbols[0]); fflush(stdout);
			}
#endif
		return;
		}
	// First create temporary tables ordered by code length
	unsigned char codelength[257];
	unsigned short codebits[256];
	int k = 0;
	int i;
	for (i = 1; i <= longest; i++) {
		for (int j = 0; j < num_symbols_of_length[i]; j++)
			codelength[k++] = i;
		}
	codelength[k] = 0;	// Sentinal
	int codeword = 0;
	k = 0;
	int total = codelength[0];
	while(codelength[k]) {
		while(codelength[k] == total) {
			codebits[k++] = codeword;
			codeword++;
			}
		codeword <<= 1;
		total++;
		}
	// And finally install in lookup table
	k = 0;
	for (int length = 1; length <= longest; length++) {
		for (i = 1; i <= num_symbols_of_length[length]; i++) {
			int entry = codebits[k] << (longest-length);
			int j = 1 << (longest-length);
#ifdef DEBUG
		if (debug) {
		    printf("%d: %d bits: ",ordered_symbols[k],codelength[k]);
		    printf("installing in table from %d to %d... ",
			entry, entry+j);
		    for (int temp = 1; temp <= codelength[k]; temp++) {
			if (codebits[k] & (1<<(codelength[k]-temp)))
				printf("1");
			else
				printf("0");
			}
		    printf("\n"); fflush(stdout);
		    }
#endif
			while(j--) {
				decode[entry] = ordered_symbols[k];
				declength[entry] = length;
				entry++;
				}
			k++;
			}
		}
	}
void compactor::put(int symbol)
{
	int cnt;
	if (!usinghuffman) {
		pfio->putbits(symbol,litlength);
#ifdef DEBUG
	if (debug)
		printf("Coder %d: uncoded symbol %d (0x%x)\n",id,symbol,symbol);
#endif
		return;
		}
	if ((cnt = enclength[symbol]) != 0)
		pfio->putbits(encode[symbol],cnt);
#ifdef DEBUG
	if (debug)
		printf("Coder %d: symbol %d (0x%x)\n",id,symbol,symbol);
#endif
	if (longest && !cnt)
		printf("Tried to code untrained symbol %d\n",symbol);
	//printf("unenc = %d, coded %d bits\n",symbol,cnt);
}

// Member functions for the rlcomp type implement a "flagword" Huffman
// coder where, if the two most frequent characters are frequent enough, the
// patterns of freqchar/nextfreq/nonfreqchar are represented in a separate
// coder of flag words.

rlcomp::rlcomp(){
		debug = 0;
		id = 0;
		inbuf = 0;
		procbuf = 0;
		symbolcnt = 0;
		psymbols = new compactor;
		pflagwords = new compactor;
		reset();
		}
rlcomp::~rlcomp() {
		delete [] inbuf;
		delete [] procbuf;
		delete psymbols;
		delete pflagwords;
		}
int rlcomp::encodelength(int maxlength){ // Called to set up encode buffers
		buflength = maxlength;
		if (!maxlength){	// This is a decoder - no buffer
			inbuf = 0;
			procbuf = 0;
			return 0;
			}
		inbuf = new unsigned char [maxlength+FLAGWORDCHUNK];
		procbuf = new unsigned char [(maxlength/FLAGWORDCHUNK)+1];
		if (!inbuf || !procbuf)
			return -1;
		inbufend = inbuf;
		procbufend = procbuf;
		readproc = procbuf;
		return 1;
		}
void rlcomp::reset() {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		for (int i = 0; i < 256; i++)
			freq[i] = 0;
		symbolcnt = 0;
		doublecode = 0;
		inbufend = inbuf; 
		procbufend = procbuf;
		readproc = procbuf;
		usingflagwords = 0; // Not really necessary. Set when building
					// or reading tree.
		psymbols->reset();
		pflagwords->reset();
#ifdef NOCOMPACTORDEBUG
			enable_compactor_debug();
#endif
		}
void rlcomp::buildtree(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		// Does major part of flagword coding
		unsigned char *readin = inbuf;
		unsigned char *writeproc = procbuf;
		unsigned char symbol;
// Debug
//		printf("inbuf:\n");
//		while(readin < inbufend)
//			printf("%x ", (int) *readin++);
//		printf("\n");
//		readin = inbuf;
// End debug
// First find the most frequent symbol
		int maxfreq = -1;
		int i;
		for (i = 0; i < 256; i++) {
			if (freq[i] > maxfreq) {
				maxfreq = freq[i];
				freqsymbol = i;
				}
			}
		int secondmaxfreq = 0;
		for (i = 0; i < 256; i++) {
			if (i == freqsymbol)
				continue;
			if (freq[i] > secondmaxfreq) {
				secondmaxfreq = freq[i];
				secondfreqsymbol = i;
				}
			}
		maxfreq += secondmaxfreq;
// Now see if it is frequent enough to justify using flagwords
// See Cheung & Keily, "An Efficient Variable Length Coding Scheme for an
// IID Source".
		if ((maxfreq < 16)||(maxfreq*2 < inbufend-inbuf)||
			(maxfreq-secondmaxfreq == inbufend-inbuf)){	
			// Too few to bother or 
			// Less than half are freqsymbol or
			// Single symbol
#ifndef NOCOMPACTORDEBUG
			if (debug)
				printf("rlcoder %d: Not using flagwords because maxfreq = %d and numchars = %d\n",id,maxfreq,(int)(inbufend-inbuf));
#endif
			usingflagwords = 0;
			// Set up symbol coder
			while(readin < inbufend)
				psymbols->inc(*readin++);
			// Don't do anything with flagword coder
			inbufend = inbuf;	// For consistency. Not necess.
#ifdef NOCOMPACTORDEBUG
			enable_compactor_debug();
#endif
			return;
			}
		usingflagwords = 1;
#ifndef NOCOMPACTORDEBUG
		if (debug) {
			printf("rlcoder %d: Using flagwords because maxfreq = %d and numchars = %d\n",id,maxfreq,(int)(inbufend-inbuf));
			printf("max freq char = %d, second max freq char = %d\n", freqsymbol, secondfreqsymbol);
			}
#endif
		int numproc = (int)(inbufend-inbuf+(FLAGWORDCHUNK-1))/FLAGWORDCHUNK;
		unsigned char *padinbufend = inbuf;
		padinbufend += (numproc*FLAGWORDCHUNK);
		for (unsigned char *p = inbufend; p < padinbufend; p++)
			*p = freqsymbol;
		int allfreq = FLAGWORDS-1;
		for (i = 0; i < numproc; i++) {
			unsigned char outbyte = 0;
			for (int j = 0; j < FLAGWORDCHUNK; j++) {
				outbyte *= 3;
				if ((symbol = *readin++) == freqsymbol)
					outbyte += 0x2;
				else if (symbol == secondfreqsymbol)
					outbyte += 0x1;
				else
					psymbols->inc(symbol);
				}
			if ((outbyte == allfreq)&&(writeproc > procbuf)&&
				(*(writeproc-1) >= allfreq)&&
				(*(writeproc-1) < MAXFLAG)) {
			// Check for two successive all frequent symbols
			// Fold into one
				pflagwords->dec(*(writeproc-1));
				(*(writeproc-1))++; // Repeat word
				pflagwords->inc(*(writeproc-1));
				}
			else {
				*writeproc++ = outbyte;
				pflagwords->inc(outbyte);
				}
			}
		procbufend = writeproc;
// Debug
//		writeproc = procbuf;
//		printf("procbuf:\n");
//		while(writeproc < procbufend)
//			printf("%x ", (int) *writeproc++);
//		printf("\n");
// End debug
		inbufend = inbuf; // Now to be used for debugging
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::writetree(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		buildtree();
		if (usingflagwords){
			putbit(1);
			int temp = (int) freqsymbol;
			int i;
			for (i = 7; i >= 0; i--) {
				putbit((temp>>i) & 1);
				}
			temp = (int) secondfreqsymbol;
			for (i = 7; i >= 0; i--) {
				putbit((temp>>i) & 1);
				}
			}
		else
			putbit(0);
		psymbols->writetree(); 
		if (usingflagwords)
			pflagwords->writetree();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::readtree(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		if (getbit()) {
			usingflagwords = 1;
			int temp = 0;
			int i;
			for (i = 0; i < 8; i++) {
				temp += getbit();
				temp <<= 1;
				}
			temp >>= 1;
			freqsymbol = (unsigned char) temp;
			temp = 0;
			for (i = 0; i < 8; i++) {
				temp += getbit();
				temp <<= 1;
				}
			temp >>= 1;
			secondfreqsymbol = (unsigned char) temp;
#ifndef NOCOMPACTORDEBUG
			if (debug)
				printf("rlcoder %d: Using flagwords, freqsymbol = %x\n",id,freqsymbol);
#endif
			symbolcnt = 0;
			doublecode = 0;
			}
		else {
#ifndef NOCOMPACTORDEBUG
			if (debug)
				printf("rlcoder %d: Not using flagwords\n",id);
#endif
			usingflagwords = 0;
			}
		psymbols->readtree(); 
		if (usingflagwords)
			pflagwords->readtree();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::maketable_encode(void){
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		psymbols->maketable_encode(); 
		if (usingflagwords)
			pflagwords->maketable_encode();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::maketable_decode(void) {
#ifdef NOCOMPACTORDEBUG
		disable_compactor_debug();
#endif
		psymbols->maketable_decode(); 
		if (usingflagwords)
			pflagwords->maketable_decode();
#ifdef NOCOMPACTORDEBUG
		enable_compactor_debug();
#endif
		}
void rlcomp::incorput(int symbol, int which) {
		if (which)
			put(symbol);
		else
			inc(symbol);
		}
void rlcomp::put(int inbyte) {
		if (!usingflagwords) {
			psymbols->put(inbyte);
			return;
			}
		if (inbyte != *inbufend++) {
			fprintf(stderr,"Internal error in rlcomp\n");
			// exit(1);
			}
		if (!(symbolcnt%FLAGWORDCHUNK)) {
			if (doublecode)
				doublecode--;
			else {
				if (*readproc >= FLAGWORDS)
					doublecode = (*readproc+1-FLAGWORDS);
				pflagwords->put(*readproc++);
				}
			}
		if ((inbyte != freqsymbol)&&(inbyte != secondfreqsymbol))
			psymbols->put(inbyte);
		symbolcnt++;
		}
