//
// Adaptive Prediction Trees
// Image Coding
//
// Version 1.0
//
// Based on...
//
// Binary tree predictive coding versions 1-5
//
// John Robinson 1994, 1995, 1997, 2002, 2003, 2004
// john@userport.com
//

// Functions for lossless palettization.
// The maximum palette supported here is 4096. The palettization algorithm is
// order N^3, so when the number of entries reaches this maximum, encoding
// time can be long.
//
// The palettization algorithm is adapted from that proposed in
// W Zeng, J Li, S Lei, "An Efficient Color Re-indexing Scheme for
// Palette-Based Compression", ICIP 2000. Warning: This scheme may be
// patented! The palettization method is therefore not a part of APT and
// is included as an illustration of how palettization may be done.
// I.e. APT supports the palettization of pictures and defines how the
// palette is encoded, but not how the palette is ordered.

#include <cstdlib>
#include <cmath>
namespace apt1_0 {
/*
//
// Temp stuff for debugging
//
int writepgm(const char *pgmname, unsigned char **pbuf,
const int& rows, const int& cols, const int& maxval)
// Write picture to "P5" pgm format
        {
        FILE *out = fopen(pgmname, "w");
        if (!out) {
                fprintf(stderr,"Cannot open %s\n", pgmname);
                return -1;
                }
        fprintf(out,"P5\n%d %d\n%d\n", cols, rows, maxval);
        for (int i = 0; i < rows; i++)
                fwrite(pbuf[i],1,cols,out);
        fclose(out);
        return 0;
        }
//
// End of temp stuff for debugging
//
*/

struct entry {
	unsigned char g;
	unsigned char r;
	unsigned char b;
	int valsum;
	int cnt;
	int index;
	entry *hist_next;
	entry *pool_prev;
	entry *pool_next;
	bool whichpool;
	};
class palettize {
	entry palette[4097];	// Extra one so can do pool-stepping without
				// memory error
	double weightsarray[4096];
	int totalentries;
	int paddedentries;	// Next highest multiple of 256 greater than
				// or equal to totalentries.
	int startindex;		// Offset of start of palette.
	int palettebits;
	int palettemask;
	entry unsorted;	// unsorted pool starts here, but this actual entry
			// is a dummy.
	entry unsorted_end;	// pool_prev points to last in unsorted pool
	entry sorted;	// sorted pool. This actual entry is a dummy.
	entry sorted_end;	// pool_prev pointer points to last in pool
	int **cooccur;	// cooccurence matrix
	int *crosscounts;	// crosscount sums
	void initpools() {
		unsorted.pool_next = palette;
		entry *pe = palette;
		pe->pool_prev = &unsorted;
		int i;
		for (i = 0; i < totalentries; i++) {
			pe->pool_next = (pe+1);
			pe->whichpool = 0;
			pe++;
			pe->pool_prev = (pe-1);
			}
		pe--;
		pe->pool_next = &unsorted_end;
		unsorted_end.pool_prev = pe;
		// Put first two entries in sorted pool
		int firsti = 0;
		int secondi = 0;
		int firstival = 0;
		int secondival = 0;
		for (i = 0; i < totalentries; i++) {
			if (crosscounts[i] > firstival) {
				firsti = i;
				firstival = crosscounts[i];
				}
			}
		for (i = 0; i < totalentries; i++) {
			if (i == firsti)
				continue;
			if (cooccur[firsti][i] > secondival) {
				secondi = i;
				secondival = cooccur[firsti][i];
				}
			}
		// printf("firsti = %d, secondi = %d\n",firsti,secondi);
		entry *pfirst = &palette[firsti];
		entry *psecond = &palette[secondi];
		// Unlink from unsorted pool
		entry *pprev = pfirst->pool_prev;
		entry *pnext = pfirst->pool_next;
		pprev->pool_next = pnext;
		pnext->pool_prev = pprev;
		pfirst->whichpool = 1;
		pprev = psecond->pool_prev;
		pnext = psecond->pool_next;
		pprev->pool_next = pnext;
		pnext->pool_prev = pprev;
		psecond->whichpool = 1;
		// Put as first and second members of sorted pool
		sorted.pool_next = pfirst;
		pfirst->pool_next = psecond;
		psecond->pool_next = &sorted_end;
		sorted_end.pool_prev = psecond;
		psecond->pool_prev = pfirst;
		pfirst->pool_prev = &sorted;
		}
	int tosorted(int index, bool frontnotback) {
		entry *pe = &palette[index];
		if (pe->whichpool)	// Already on sorted list
			return -1;
		entry *pprev = pe->pool_prev;
		entry *pnext = pe->pool_next;
		// Unlink from unsorted pool
		pprev->pool_next = pnext;
		pnext->pool_prev = pprev;
		pe->whichpool = 1;
		// Link to front or back or sorted pool
		if (frontnotback) {
			pe->pool_next = sorted.pool_next;
			sorted.pool_next = pe;
			pe->pool_next->pool_prev = pe;
			pe->pool_prev = &sorted;
			}
		else {
			pe->pool_prev = sorted_end.pool_prev;
			sorted_end.pool_prev = pe;
			pe->pool_prev->pool_next = pe;
			pe->pool_next = &sorted_end;
			}
		return 0;
		}
		
	void createcoocur(unsigned char ***components, int nr, int nc,
					entry *hist[256][256]) {
		cooccur = new int *[totalentries];
		int i;
		for (i = 0; i < totalentries; i++)
			cooccur[i] = new int[totalentries];
		crosscounts = new int[totalentries];
		for (i = 0; i < totalentries; i++)
		for (int j = 0; j < totalentries; j++)
			cooccur[i][j] = 0;
		for (i = 0; i < totalentries; i++)
			crosscounts[i] = 0;
		// Just do 1D for now
		for (i = 0; i < nr; i++) {
			unsigned char *g = components[0][i];
			unsigned char *r = components[1][i];
			unsigned char *b = components[2][i];
			int prev = -1;	// To skip first round
			for (int j = 0; j < nc; j++) {
				entry *p = hist[*g][*r];
				while (p) {
					if (p->b == *b)
						break;
					p = p->hist_next;
					}
				g++; r++; b++;
				if (prev >= 0) {
					cooccur[prev][p->index]++;
					cooccur[p->index][prev]++;
					}
				prev = p->index;
				}
			}
		for (i = 0; i < totalentries; i++)
		for (int j = 0; j < totalentries; j++) {
			if (j == i)
				continue;
			crosscounts[i] += cooccur[i][j];
			}
		}

	void deletecooccur() {
		for (int i = 0; i < totalentries; i++)
			delete [] cooccur[i];
		delete [] cooccur;
		delete [] crosscounts;
		}
	void initweights() {
		double log2 = log(2.0);
		for (int i = 0; i < totalentries; i++) {
			weightsarray[i] = log(1.0+1.0/(1+(double)i));
			weightsarray[i] /= log2; // Necessary?
			}
		}
	int move_next_to_sorted_pool() {
		entry *src = unsorted.pool_next;
		if (src == &unsorted_end)
			return 0;
		entry *bestleft = src;
		entry *bestright = src;
		double bestleftval = 0;
		double bestrightval = 0;
		int indexarray[4096];
		int numsorted = 0;
		// Set up indices from sorted pool
		entry *p = sorted.pool_next;
		while (p != &sorted_end) {
			indexarray[numsorted++] = p->index;
			p = p->pool_next;
			}
		while (src != &unsorted_end) { // Main loop
			double left = 0;
			double right = 0;
			for (int i = 0; i < numsorted; i++) {
				double curr = 
				  (double) cooccur[src->index][indexarray[i]];
				left += curr*weightsarray[i];
				right += curr*weightsarray[numsorted-1-i];
				}
			if (left > bestleftval) {
				bestleftval = left;
				bestleft = src;
				}
			if (right > bestrightval) {
				bestrightval = right;
				bestright = src;
				}
			src = src->pool_next;
			}
		if (bestrightval > bestleftval) 
			tosorted(bestright->index,0);
		else
			tosorted(bestleft->index,1);
		return 1;
		}
/*
// Debug function
	void print_pools() {
		entry *pe = unsorted.pool_next;
		printf("Unsorted pool:\n");
		while (pe != &unsorted_end) {
			printf("\tIndex %d: (%d,%d,%d), cnt = %d\n",
				pe->index, pe->r, pe->g, pe->b, pe->cnt);
			pe = pe->pool_next;
			}
		pe = sorted.pool_next;
		int cnt = 0;
		printf("Sorted pool:\n");
		while (pe != &sorted_end) {
			printf("\t (%d) Index %d: (%d,%d,%d), cnt = %d\n",
				cnt, pe->index, pe->r, pe->g, pe->b, pe->cnt);
			pe = pe->pool_next;
			cnt++;
			}
		}
*/
public:
	// Most of the actual palettization is done in the private functions.
	// They are called by the constructor below.
	palettize(unsigned char ***components, int nr, int nc, int quality,
							bool& cando) {
		int maxentries = 10;
		if (quality > 70)
			maxentries = 256;
		if ((quality == 100)&&(nr*nc > 512*512))
		// Size criterion is because palette will take significant
		// extra space. Over a broad range around 0.25Mbytes, graphics
		// are about equally well coded with and without palettes.
			maxentries = 4096 - 256;
		entry *histogram[256][256];
		int i, j;
		for (i = 0; i < 256; i++)
		for (j = 0; j < 256; j++)
			histogram[i][j] = 0;
		for (i = 0; i < maxentries; i++)
			palette[i].cnt = 0;
		for (i = 0; i < nr; i++) {
			unsigned char *g = components[0][i];
			unsigned char *r = components[1][i];
			unsigned char *b = components[2][i];
			for (j = 0; j < nc; j++) {
				entry *p = histogram[*g][*r];
				while(p) {
					if (p->b == *b) {
						p->cnt++;
						break;
						}
					p = p->hist_next;
					}
				if (!p) {
					entry *q = new entry;
					q->g = *g;
					q->r = *r;
					q->b = *b;
					q->cnt = 1;
					q->hist_next = histogram[*g][*r];
					histogram[*g][*r] = q;
					}
				g++; r++; b++;
				}
			}
		totalentries = 0;
		int paletteindex = 0;
		for (i = 0; i < 256; i++) {
	    		for (j = 0; j < 256; j++) {
				int numentries = 0;
				entry *p = histogram[i][j];
				while(p) {
					numentries++;
					p = p->hist_next;
					}
				totalentries += numentries;
				if (totalentries > maxentries) {
					// printf("Too many to palettize\n");
					cando = 0;
					return;
					}
				p = histogram[i][j];
				while(p) {
					p->index = paletteindex;
					palette[paletteindex] = *p;
					palette[paletteindex++].hist_next = p;
					// In the palette the hist_next pointers
					// point back to the histogram
					p = p->hist_next;
					}
				}
	    		}
		// printf("Palette size = %d\n", totalentries);
		palettebits = 0;
		while (totalentries > (1<<palettebits))
			palettebits++;
		// palettebits++;
		// printf("Palette bits = %d\n", palettebits);
		palettemask = (1<<palettebits) - 1;
		paddedentries = (totalentries>>8)<<8;
		if (totalentries > paddedentries) {
			paddedentries += 256;
			if (!(paddedentries & 0x100)) // Don't want even number
				paddedentries = totalentries; // of 256-entry blocks
			}
		cando = 1;
		// Reorganise palette for more efficient coding
		createcoocur(components, nr, nc, histogram);
		initpools();
		// print_pools();
		initweights();
		while(move_next_to_sorted_pool());
		// print_pools();
		deletecooccur();
		// So now the ordering algorithm is done, we just have
		// to get the data structures right.
		// Work out the start index
		startindex = (paddedentries - totalentries)/2; // For now
		// Reorder index values in the palette and copy in order
		// to temp palette
		entry *temp = new entry[paddedentries];
		entry *pe = sorted.pool_next;
		int newindex = startindex;
		while(pe != &sorted_end) {
			pe->index = newindex;
			temp[newindex++] = *pe;
			pe = pe->pool_next;
			}
		// Overwrite palette with ordered palette
		// (pointers will no longer be valid)
		for (i = startindex; i < totalentries+startindex; i++) 
			palette[i] = temp[i];
		delete [] temp;
		// Reorder histogram index values;
		pe = palette+startindex;
		for (i = 0; i < totalentries; i++) {
			pe->hist_next->index = pe->index;
			pe++;
			}
/*
		printf("First palette value = %d,%d,%d. cnt = %d\n",palette[0].r,palette[0].g, palette[0].b, palette[0].cnt);
		printf("255 palette value = %d,%d,%d. cnt = %d\n",palette[255].r,palette[255].g, palette[255].b, palette[255].cnt);
		printf("256 palette value = %d,%d,%d. cnt = %d\n",palette[256].r,palette[256].g, palette[256].b, palette[256].cnt);
		printf("Last palette value = %d,%d,%d. cnt = %d\n",palette[totalentries-1].r,palette[totalentries-1].g, palette[totalentries-1].b, palette[totalentries-1].cnt);
*/


		// Now actually do the palette mapping
		for (i = 0; i < nr; i++) {
			unsigned char *g = components[0][i];
			unsigned char *r = components[1][i];
			unsigned char *b = components[2][i];
			for (j = 0; j < nc; j++) {
				entry *p = histogram[*g][*r];
				while(p) {
					if (p->b == *b) {
						break;
						}
					p = p->hist_next;
					}
				if (!p) {
					printf("Internal error in palettize\n");
					return;
					}
				if (palettebits > 8) {
				//	*g = ((p->index>>(palettebits-8)))&0xff;
				//	*r = p->index & (palettemask>>8);
				// Note there is one last trick in here. We
				// "ripple" the coding of the least significant
				// eight bits so that boundaries between
				// 256-entry blocks are not abrupt.
					int topbits = *r = (p->index>>8)&0xff;
					if (topbits & 1)
						*g = 255 - (p->index&0xff);
					else
						*g = p->index&0xff;
					}
				else {
					*g = p->index;
					*r = 0;
					}
				*b = 0;
				g++; r++; b++;
				}
			}
/*
		// Debug stuff
		writepgm("palpic1", components[0], nr, nc, 255);
		if (palettebits > 8)
			writepgm("palpic2", components[1], nr, nc, 15);
*/
		// Clean up
		for (i = 0; i < 256; i++)
		for (j = 0; j < 256; j++) {
			entry *p = histogram[i][j];
			entry *q;
			while(p) {
				q = p->hist_next;
				delete p;
				p = q;
				}
			}
		}
	~palettize() {
		}
	// Decoder calls frompalette to restore actual values back
	int frompalette(unsigned char ***components, int nr, int nc) {
		int i, j;
		for (i = 0; i < nr; i++) {
			unsigned char *g = components[0][i];
			unsigned char *r = components[1][i];
			unsigned char *b = components[2][i];
			for (j = 0; j < nc; j++) {
				int index = *g;
				if (palettebits > 8) {
					// Reverse the ripple!
					int topbits = *r;
					if (topbits & 1)
						index = 255 - index;
					index |= (topbits << 8);
					}
					// index = (index<<(palettebits-8))+*r;
				*g++ = palette[index].g;
				*r++ = palette[index].r;
				*b++ = palette[index].b;
				}
			}
		return 0;
		}
	void writepalette(FILE *cfp, cfileio& fio) {
		rlcomp codeout;
		codeout.encodelength(1024*64);
		codeout.attach(&fio);
		codeout.indexnumber(0);
		// codeout.debugon();
		fputc((unsigned char) ((totalentries>>8)&0xff), cfp);
		fputc((unsigned char) (totalentries&0xff), cfp);
		fputc((unsigned char) startindex, cfp);
/**/
		unsigned char prevval = 0;
		int i;
		for (i = startindex; i < totalentries+startindex; i++) {
			palette[i].g = palette[i].g - prevval;
			prevval = palette[i].g + prevval;
			}
		prevval = 0;
		for (i = startindex; i < totalentries+startindex; i++) {
			palette[i].r = palette[i].r - prevval;
			prevval = palette[i].r + prevval;
			}
		prevval = 0;
		for (i = startindex; i < totalentries+startindex; i++) {
			palette[i].b = palette[i].b - prevval;
			prevval = palette[i].b + prevval;
			}
/**/
		for (i = startindex; i < totalentries+startindex; i++)
			codeout.inc(palette[i].g);
		codeout.writetree();
		for (i = startindex; i < totalentries+startindex; i++)
			codeout.put(palette[i].g);
		codeout.reset();
		for (i = startindex; i < totalentries+startindex; i++)
			codeout.inc(palette[i].r);
		codeout.writetree();
		for (i = startindex; i < totalentries+startindex; i++)
			codeout.put(palette[i].r);
		codeout.reset();
		for (i = startindex; i < totalentries+startindex; i++)
			codeout.inc(palette[i].b);
		codeout.writetree();
		for (i = startindex; i < totalentries+startindex; i++)
			codeout.put(palette[i].b);
		fio.flush();
		}
	// Constructor below is used by decoder and reads the palette in
	palettize(FILE *cfp, cfileio& fio) {
		rlcomp codein;
		codein.attach(&fio);
		codein.indexnumber(0);
		// codein.debugon();
		unsigned char topbyte, bottombyte;
		topbyte = fgetc(cfp);
		bottombyte = fgetc(cfp);
		startindex = fgetc(cfp);
		totalentries = (topbyte<<8) + bottombyte;
//		totalentries = fgetc(cfp);
//		if (!totalentries)
//			totalentries = 256;
		// printf("Palette size = %d\n", totalentries);
		palettebits = 0;
		while (totalentries > (1<<palettebits))
			palettebits++;
		// palettebits++;
		palettemask = (1<<palettebits) - 1;
		paddedentries = (totalentries>>8)<<8;
		if (totalentries > paddedentries) {
			paddedentries += 256;
			if (!(paddedentries & 0x100)) // Don't want even number
				paddedentries = totalentries; // of 256-entry blocks
			}
		codein.readtree();
		int i;
		for (i = startindex; i < totalentries+startindex; i++) 
			palette[i].g = codein.get();
		codein.readtree();
		for (i = startindex; i < totalentries+startindex; i++) 
			palette[i].r = codein.get();
		codein.readtree();
		for (i = startindex; i < totalentries+startindex; i++) 
			palette[i].b = codein.get();
		fio.rewind();
/**/
		unsigned char prevval = 0;
		for (i = startindex; i < totalentries+startindex; i++)
			prevval = palette[i].g = palette[i].g + prevval;
		prevval = 0;
		for (i = startindex; i < totalentries+startindex; i++)
			prevval = palette[i].r = palette[i].r + prevval;
		prevval = 0;
		for (i = startindex; i < totalentries+startindex; i++)
			prevval = palette[i].b = palette[i].b + prevval;
/**/
		}
		
	};

class bwpalettize {
	unsigned char palette[10];
	int reversepalette[256];
	unsigned char *p;
	int totalentries;
	int palettebits;
	int palettemask;
public:
	bwpalettize(unsigned char ***components, int nr, int nc, int quality,
							bool& cando) {
		int i, j;
		totalentries = 0;
		for (i = 0; i < 256; i++)
			reversepalette[i] = -1;
		for (i = 0; i < nr; i++) {
			p = components[0][i];
			for (j = 0; j < nc; j++) {
				unsigned char pel = *p++;
				if (reversepalette[pel] < 0) {
					if (totalentries == 10) {
						cando = 0;
						return;
						}
					reversepalette[pel] = totalentries;
					palette[totalentries++] = pel;
					}
				}
			}
		// printf("Palette size = %d\n", totalentries);
		palettebits = 0;
		while (totalentries > (1<<palettebits))
			palettebits++;
		palettemask = (1<<palettebits) - 1;
		cando = 1;

		// Now actually do the palette mapping
		for (i = 0; i < nr; i++) {
			p = components[0][i];
			for (j = 0; j < nc; j++) {
				*p = (unsigned char) reversepalette[*p];
				p++;
				}
			}
		}
	~bwpalettize() {
		}
	int frompalette(unsigned char ***components, int nr, int nc) {
		int i, j;
		for (i = 0; i < nr; i++) {
			unsigned char *p = components[0][i];
			for (j = 0; j < nc; j++) {
				*p = palette[*p];
				p++;
				}
			}
		return 0;
		}
	void writepalette(FILE *cfp, cfileio& fio) {
		rlcomp codeout;
		codeout.encodelength(1024*64);
		codeout.attach(&fio);
		codeout.indexnumber(0);
		// codeout.debugon();
		fputc((unsigned char) (totalentries&0xff), cfp);
		unsigned char prevval = 0;
		int i;
		for (i = 0; i < totalentries; i++) {
			palette[i] = palette[i] - prevval;
			prevval = palette[i] + prevval;
			}
		for (i = 0; i < totalentries; i++)
			codeout.inc(palette[i]);
		codeout.writetree();
		for (i = 0; i < totalentries; i++)
			codeout.put(palette[i]);
		fio.flush();
		}
	bwpalettize(FILE *cfp, cfileio& fio) {
		rlcomp codein;
		codein.attach(&fio);
		codein.indexnumber(0);
		// codein.debugon();
		totalentries = fgetc(cfp);
		if (!totalentries)
			totalentries = 256;
		// printf("Palette size = %d\n", totalentries);
		palettebits = 0;
		while (totalentries > (1<<palettebits))
			palettebits++;
		// palettebits++;
		palettemask = (1<<palettebits) - 1;
		codein.readtree();
		int i;
		for (i = 0; i < totalentries; i++) 
			palette[i] = codein.get();
		fio.rewind();
		unsigned char prevval = 0;
		for (i = 0; i < totalentries; i++)
			prevval = palette[i] = palette[i] + prevval;
		}
		
	};
};
